local _, private = ...;
local Core = private.Core;
local Gui = private.Gui;
local Settings = private.Settings;

local function SlashCmd (...)
  Gui.SlashCmd(...);
end

local intervals = {};

local defaults = {
  profile = {
    intervals = {}
  }
};

local function FindInterval (intervalList, targetPercent)
  local closestPercent = 0;
  local closestSound = LibStub("LibSharedMedia-3.0"):GetDefault("sound");

  for percent, sound in pairs(intervalList) do
    if percent >= closestPercent and percent <= targetPercent then
      closestPercent = percent;
      closestSound = sound;
    end
  end

  return closestSound, closestPercent;
end

local function CountIntervals (spellName)
  local intervalList = intervals[spellName];
  
  if intervalList == nil then
    return 0;
  end
  
  local count = 0;
  for k,v in pairs(intervalList) do
    count = count + 1;
  end
  return count;
end

function Core.GetInterval (spellName, percent)
  local intervalList = intervals[spellName];
  if intervalList then
    return FindInterval(intervalList, percent);
  else
    return 0, LibStub("LibSharedMedia-3.0"):GetDefault("sound");
  end
end

function Core.SetInterval (spellName, percent, sound)
  local intervalList = intervals[spellName] or {};
  intervals[spellName] = intervalList;

  local oldSound, oldPercent = FindInterval(intervals[spellName], percent);
  intervalList[percent] = sound;

  if oldPercent == percent then
    dinger.callbacks:Fire("DINGER_REMOVE_INTERVAL", percent, oldSound);
  end

  dinger.callbacks:Fire("DINGER_NEW_INTERVAL", percent, sound);
end

function Core.RemoveInterval (spellName, percent)
  local intervalList = intervals[spellName];
  
  if nil == intervalList then
    return;
  end

  local _, truePercent = FindInterval(intervalList, percent);

  if 0 == truePercent then
    local sound = LibStub("LibSharedMedia-3.0"):GetDefault("sound");
    Core.SetInterval(spellName, truePercent, sound);
  else
    local oldSound = assert(intervalList[truePercent]);
    intervalList[truePercent] = nil;
    dinger.callbacks:Fire("DINGER_REMOVE_INTERVAL", truePercent, oldSound);
  end
end

local defaultTable = {
  [0] = "None",
}

function Core.SpellHasIntervals (spellName)
  local intervalList = intervals[spellName];
  if nil == intervalList then
    return false;
  end

  if intervalList[0] ~= "None" then
    return true;
  end

  if CountIntervals(spellName) > 1 then
    return true;
  end
  
  return false;
end

function Core.Intervals (spellName)
  if CountIntervals(spellName) == 0 then
    return next, defaultTable, nil;
  else
    return next, intervals[spellName], nil;
  end
end

function dinger:OnInitialize ()
  dinger.callbacks = dinger.callbacks or
    LibStub("CallbackHandler-1.0"):New(dinger);

  db = LibStub("AceDB-3.0"):New("dingerDB", defaults);

  intervals = db.profile.intervals;

  for soundName, soundPath in pairs(private.Sounds.files) do
    LibStub("LibSharedMedia-3.0"):Register("sound", soundName, soundPath);
  end
  
  Gui.Init();

  SLASH_DINGER1 = "/dinger";
  SlashCmdList["DINGER"] = SlashCmd;

  dinger.debug = true;
  self:RegisterEvent("UNIT_SPELLCAST_FAILED");  
end

local counts = {};

function dinger:UNIT_SPELLCAST_FAILED (arg1, unit, spell)
  local start, duration, enable = GetSpellCooldown(spell);

  -- item cooldowns
  if nil == start then
    return;
  end

  if unit == "player" and start > 0 and enable == 1 then
    local fin = start + duration;
    local now = GetTime();
    local spent = now - start; 
    local left = math.floor((fin - now) + 0.5);
    local percent = spent/duration;

    local sound, interval = Core.GetInterval(spell, percent);
    local sound_file = LibStub("LibSharedMedia-3.0"):Fetch("sound", sound);
    if sound_file then
      PlaySoundFile(sound_file);
    end  
  end
end

--------------------------------------------------------------------------------
-- PUBLIC API
--
-- dinger's unit of data is an "interval." When the player tries to activate an
-- ability and fails due to a cooldown, dinger looks to see how far along that
-- cooldown is. The cooldown's progress determines what spell (if any) dinger
-- plays.
--
-- Intervals are identified by the percent of completion at which they can first
-- be triggered. For example, if an interval starts at 0.5, then its sound will
-- be played when the cooldown is half over or more and the user tries to
-- activate the corresponding spell.
--
-- It's a little more complicated than that, actually. Intervals don't have
-- predefined endpoints - so if you have an interval with percent 0.5, it could
-- responible for any hit in the range [0.5, 0.6], or it could be [0.5, 0.8],
-- or even [0.5, 1]. An intervals endpoint is determined by what intervals
-- come after it; if none do, it's responsible for all sounds until the cooldown
-- completes.
--
-- It's important to note that all spells always have at least one interval
-- assigned to them, even if that interval is [0, 1] and maps to a silent sound
-- file.
--------------------------------------------------------------------------------

-- Iterates over the intervals for a given spell.
function dinger:Intervals (spellName)
  return Core.Intervals(spellName);
end

-- Returns the interval associated with 'percent'. 'spellName' is the name of the
-- spell, 'percent' is the percent of the interval you're interested in. If
-- 'percent' doesn't land exactly on an interval's start point, it returns the
-- interval owning 'percent'.
function dinger:GetInterval (spellName, percent)
  return Core.GetInterval(spellName, percent);
end

-- Sets an interval for a spell. spellName is the name of the spell, percent
-- is the start of the interval, and sound is the sound handle.
function dinger:SetInterval (spellName, percent, sound)
  return Core.SetInterval(spellName, percent, sound);
end

-- Removes an interval. spellName is the name of the spell, and percent is
-- the percent of the interval to remove. If percent doesn't match an interval,
-- the interval responsible for that percent is removed instead. For example,
-- if you have these intervals for SomeSpell:
--   [0, 0.5] = sound1
--   [0.5, 1] = sound2
-- then dinger:RemoveInterval("SomeSpell", 0.7) will remove the second
-- interval.
--
-- You can never remove the 0 interval. Doing so will reset its sound to the
-- default.
function dinger:RemoveInterval (spellName, percent)
  return Core.RemoveInterval(spellName, percent);
end