local _, private = ...;
local Gui = private.Gui;
local Bar = private.Gui.Bar;
local Core = private.Core;
local Settings = private.Settings;
local DropDown = private.Gui.DropDown;

local intervalsByButton = {};

local function GetCursorX (left)
  local x = GetCursorPosition() / UIParent:GetScale();
  local deltaX = x - left;
  
  return deltaX;
end

local function Interval_Update (intervalButton)
  local deltaX = GetCursorX(intervalButton:GetLeft());
  
  dingerBarFollower:ClearAllPoints();
  dingerBarFollower:SetPoint("CENTER", intervalButton, "LEFT", deltaX, 0)
end

local function Interval_OnEnter (intervalButton)
  local self = intervalsByButton[intervalButton];
  self.button:SetScript("OnUpdate", Interval_Update);
  dingerBarFollower:Show();
end

local function Interval_OnLeave (intervalButton)
  local self = intervalsByButton[intervalButton];
  self.button:SetScript("OnUpdate", nil);
  dingerBarFollower:Hide();
end

local function Interval_OnClick (intervalButton, mouseButton)
  PlaySound("igMainMenuOptionCheckBoxOn");
  local interval = intervalsByButton[intervalButton];
  interval["On" .. tostring(mouseButton) .. "Click"](interval);
end

local function SoundButton_OnClick (soundButton, mouseButton)
  if mouseButton == "LeftButton" then
    local sound = LibStub("LibSharedMedia-3.0"):Fetch(
      "sound", soundButton.interval.sound);
    PlaySoundFile(sound);
  else
    soundButton.interval:OnRightButtonClick();
  end
end

local DIVIDER_HEIGHT = 32;
local DIVIDER_WIDTH = DIVIDER_HEIGHT;

local DIVIDER_NORMAL_TEXTURE =
  "Interface\\AddOns\\dinger\\art\\DoubleEndedTick.tga";
local DIVIDER_HIGHLIGHT_TEXTURE =
  "Interface\\AddOns\\dinger\\art\\DoubleEndedTickHighlight.tga";

local SOUNDBUTTON_NORMAL_TEXTURE =
  "Interface\\COMMON\\VOICECHAT-SPEAKER.png";
local SOUNDBUTTON_HIGHLIGHT_TEXTURE =
  "Interface\\COMMON\\VOICECHAT-ON.png";
local SOUNDBUTTON_DISABLED_TEXTURE =
  "Interface\\COMMON\\VOICECHAT-MUTED.png";

local dingerBarFollowerTable = {
  name = "dingerBarFollower",
  type = "frame",
  height = DIVIDER_HEIGHT,
  width = DIVIDER_WIDTH,
  frameStrata = "HIGH",
  frameLevel = 3,
  children = {
    {
      name = "$parentTexture",
      type = "texture",
      drawLayer = "ARTWORK",
      texture = DIVIDER_NORMAL_TEXTURE,
      color = {
        r = 1.0,
        g = 1.0,
        b = 1.0,
        a = 0.6,
      },
      anchors = {
        allPoints = true,
      },
    },
  },
};

local BAR_HEIGHT = 24;
local BAR_WIDTH = BAR_HEIGHT * 8;

local FILL_FULL = {
  r = 0.35,
  g = 0.35,
  b = 0.5,
  a = 0.8,
};

local FILL_PARTIAL = {
  r = 0.9,
  g = 0.9,
  b = 1.0,
  a = 0.8,
};

local FILL_INACTIVE = {
  r = 1,
  g = 1,
  b = 0.3,
  a = 0.8,
};

local function Bar_OnUpdate (...)
  Bar.OnUpdate(...);
end

local dingerBarTable = {
  name = "dingerBar",
  type = "button",
  frameStrata = "HIGH",
  frameLevel = 2,
  height = BAR_HEIGHT,
  width = BAR_WIDTH,
  anchors = {
    BOTTOM = {
      relativeTo = "!$parent",
      xOff = 0,
      yOff = 45,
    },
  },
  scripts = {
    OnUpdate = Bar_OnUpdate,
  },
  children = {
    {
      name = "$parentBackgroundTexture",
      type = "texture",
      drawLayer = "BACKGROUND",
      texture = "Interface\\GLUES\\LoadingBar\\Loading-BarBackground.png",
      anchors = {
        allPoints = true,
      },
    },
    {
      name = "$parentBorderTexture",
      type = "texture",
      drawLayer = "OVERLAY",
      texture = "Interface\\GLUES\\LoadingBar\\Loading-BarBorder.png",
      anchors = {
        TOPLEFT = {
          relativeTo = "!$parent",
          relativePoint = "TOPLEFT",
          xOff = -14,
          yOff = 16,
        },
        TOPRIGHT = {
          relativeTo = "!$parent",
          relativePoint = "TOPRIGHT",
          xOff = 14,
          yOff = 16,
        },
        BOTTOMLEFT = {
          point = "BOTTOMLEFT",
          relativeTo = "!$parent",
          relativePoint = "BOTTOMLEFT",
          xOff = -14,
          yOff = -16,
        },
        BOTTOMRIGHT = {
          point = "BOTTOMRIGHT",
          relativeTo = "!$parent",
          relativePoint = "BOTTOMRIGHT",
          xOff = 14,
          yOff = -16,
        },
      },
    },
    {
      name = "$parentFillTexture",
      type = "texture",
      drawLayer = "ARTWORK",
      color = {
        r = 0.3,
        g = 0.3,
        b = 1.0,
        a = 0.8,
      },
      anchors = {
        TOPLEFT = {},
        BOTTOMLEFT = {},
        RIGHT = {},
      },
    },
    dingerBarFollowerTable,
  },
};

local dividerButtonTable = {
  type = "button",
  height = DIVIDER_HEIGHT,
  width = DIVIDER_WIDTH,
  normalTexture = DIVIDER_NORMAL_TEXTURE,
  highlightTexture = DIVIDER_HIGHLIGHT_TEXTURE,
};

local intervalButtonTable = {
  type = "button",
  height = 20,
  registerForClicks = {
    "LeftButtonUp",
    "RightButtonUp",
  },
  scripts = {
    OnEnter = Interval_OnEnter,
    OnLeave = Interval_OnLeave,
    OnClick = Interval_OnClick,
  },
  children = {
    {
      name = "$parentSoundButton",
      type = "button",
      height = 20,
      width = 20,
      normalTexture = SOUNDBUTTON_NORMAL_TEXTURE,
      highlightTexture = SOUNDBUTTON_HIGHLIGHT_TEXTURE,
      registerForClicks = {
        "LeftButtonUp",
        "RightButtonUp",
      },
      scripts = {
        OnClick = SoundButton_OnClick,
      },
      anchors = {
        BOTTOM = {
          relativeTo = "!$parent",
          relativePoint = "CENTER",
          xOff = 0,
          yOff = 25,
        },
      },
    },
  },
};

local Divider, Interval;

local intervalListHead;

function Bar.Init (parent)
  Bar.frame = LibStub("LuaToGui"):Create(dingerBarTable, parent);
  Bar.cursorFollower = dingerBarFollower;
end

function Bar.DisplaySpell (spellName)
  Bar.EraseDisplay();
  if spellName then
    local interval = Interval:New(0, 1);
    intervalListHead = interval;
    interval:SetBetween();
    
    for percent, sound in dinger:Intervals(spellName) do
      Bar.AddIntervalAtPercent(percent, sound);
    end
  end
end

function Bar.EraseDisplay ()
  while intervalListHead do
    local interval = intervalListHead;
    intervalListHead = interval.next;
    interval:Free();
  end
end

function Bar.OnUpdate ()
  local spell = Gui.GetCurrentSpell();
  local color = FILL_INACTIVE;
  
  if spell then
    local start, duration = GetSpellCooldown(spell);
    
    if duration ~= null and duration > 1.5 then
      local timePassed = GetTime() - start;
      local percentDone = timePassed / duration;
      local barWidth = dingerBar:GetRight() - dingerBar:GetLeft();
      local xOff =  barWidth * percentDone;
      dingerBarFillTexture:SetPoint("RIGHT", dingerBar, "LEFT", xOff, 0);
      color = FILL_PARTIAL;
    else
      color = FILL_FULL;
    end
  end

  if color ~= FILL_PARTIAL then
    dingerBarFillTexture:SetPoint("RIGHT");
  end
  
  dingerBarFillTexture:SetVertexColor(color.r, color.g, color.b, color.a);
end

function Bar.GetActiveInterval ()
  return Bar.activeInterval;
end

function Bar.SetActiveInterval (interval)
  Bar.activeInterval = interval;
end

local function findIntervalByPercent (percent)
  local interval = intervalListHead;
  while interval do
    -- bingo?
    if percent >= interval.startP and percent < interval.endP then
      break;
    else
      interval = interval.next;
    end
  end
  
  return interval;
end

function Bar.AddIntervalAtPercent (percent, sound)
  local interval;
  
  if 0 == percent then
    interval = intervalListHead;
  else
    local oldInterval = findIntervalByPercent(percent);
    interval = oldInterval:Split(percent);
  end
  
  interval.sound = sound;
  
  local texture, highlightTexture;
  if interval.sound == LibStub("LibSharedMedia-3.0"):GetDefault("sound") then
    texture = SOUNDBUTTON_DISABLED_TEXTURE;
    highlightTexture = nil;
  else
    texture = SOUNDBUTTON_NORMAL_TEXTURE;
    highlightTexture = SOUNDBUTTON_HIGHLIGHT_TEXTURE;
  end

  local soundButton = _G[interval.button:GetName() .. "SoundButton"];
  soundButton:SetNormalTexture(texture);
  soundButton:SetHighlightTexture(highlightTexture, "ADD");
end

function Bar.RemoveIntervalAtPercent (percent)
  local oldInterval = findIntervalByPercent(percent);
  if 0 == oldInterval.startP then
    oldInterval.sound = LibStub("LibSharedMedia-3.0"):GetDefault("sound");
  else  
    oldInterval:Free();
  end
end

function Bar.Clear ()
  while intervalListHead.next do
    Core.RemoveInterval(Gui.GetCurrentSpell(), intervalListHead.next.startP);
  end
  Core.RemoveInterval(Gui.GetCurrentSpell(), intervalListHead.startP);
end

Interval = {};

local intervalButtonCount = 1;

local function CreateIntervalButton ()
  intervalButtonTable.name = "$parentIntervalButton" .. intervalButtonCount;
  intervalButtonCount = intervalButtonCount + 1;
  local button = LibStub("LuaToGui"):Create(intervalButtonTable,
                                            dingerBar);
  return button;
end



local dividerButtonCount = 1;

local function CreateDividerButton ()
  dividerButtonTable.name = "$parentDividerButton" .. dividerButtonCount;
  dividerButtonCount = dividerButtonCount + 1;
  return LibStub("LuaToGui"):Create(dividerButtonTable, dingerBar);
end

local dividerButtonRecycler = Recycler:New(CreateDividerButton);
local intervalButtonRecycler = Recycler:New(CreateIntervalButton);

function Interval:New (startP, endP)
  local interval = setmetatable({}, { __index = Interval });
  interval.button = intervalButtonRecycler:Get();
  interval.button:Show();
  _G[interval.button:GetName() .. "SoundButton"].interval = interval;
  interval.startP = startP;
  interval.endP = endP;
  intervalsByButton[interval.button] = interval;
  return interval;
end

function Interval:SetBetween (leftDivider, rightDivider)
  self.leftDivider = leftDivider;
  self.rightDivider = rightDivider;
  
  if not leftDivider then
    self.button:SetPoint("TOPLEFT");
    self.button:SetPoint("BOTTOMLEFT");
  else
    self.button:SetPoint("TOPLEFT", leftDivider.button, "TOPRIGHT");
    self.button:SetPoint("BOTTOMLEFT", leftDivider.button, "BOTTOMRIGHT");
  end
  
  if not rightDivider then
    self.button:SetPoint("TOPRIGHT");
    self.button:SetPoint("BOTTOMRIGHT");
  else
    self.button:SetPoint("TOPRIGHT", rightDivider.button, "TOPLEFT");
    self.button:SetPoint("BOTTOMRIGHT", rightDivider.button, "BOTTOMLEFT");
  end
end

function Interval:Split (intervalPercent)
  local newInterval = Interval:New(intervalPercent, self.endP);
  self.endP = intervalPercent;
  
  local newDivider = Divider:New(intervalPercent);
  newInterval:SetBetween(newDivider, self.rightDivider);
  self:SetBetween(self.leftDivider, newDivider);

  if self.next then
    self.next.prev = newInterval;
  end

  newInterval.next = self.next;
  newInterval.prev = self;
  self.next = newInterval;

  return newInterval;
end

function Interval:OnLeftButtonClick ()
  local left = dingerBar:GetLeft();
  local right = dingerBar:GetRight();
  
  local maxDelta = right - left;

  local deltaX = GetCursorX(left);
  local percentDelta = deltaX / maxDelta;

  local intervalSplitting = findIntervalByPercent(percentDelta);
  Core.SetInterval(Gui.GetCurrentSpell(), percentDelta, intervalSplitting.sound);
end

function Interval:OnRightButtonClick ()
  Bar.SetActiveInterval(self);
  ToggleDropDownMenu(1, nil, DropDown.getFrame(), self.button:GetName(), 0, 0);
end

function Interval:Free ()
  if self.prev then
    self.prev:SetBetween(self.prev.leftDivider, self.rightDivider);
    self.prev.endP = self.endP;
    self.prev.next = self.next;
  end
  
  if self.next then
    self.next.prev = self.prev;
  end

  if self.leftDivider then
    self.leftDivider:Free();
  end

  if intervalListHead == self then
    intervalListHead = self.next;
  end
  
  self.button:Hide();  
  intervalButtonRecycler:Free(self.button);  
end

Divider = {};

function Divider:New (percent)
  local divider = setmetatable({}, { __index = Divider });
  divider.button = dividerButtonRecycler:Get();
  divider.button:Show();
  divider:Place(percent);
  return divider;
end

function Divider:Free ()
  self.button:Hide()
  dividerButtonRecycler:Free(self.button);
end

function Divider:Place (percent)
  self.percent = percent;
  local barLength = dingerBar:GetRight() - dingerBar:GetLeft();
  local x = percent * barLength;
  self.button:SetPoint("CENTER", dingerBar, "LEFT", x, 0);
end